import logging
from contextlib import ContextDecorator
from logging import Logger

import catcher
from catcher.modules.log_storage import EmptyLogStorage
from colorama import Fore, Style

log_storage = EmptyLogStorage('empty')
colored_output = True  # use colorama for colored output. Is controlled by --no-color cmd arg
output_enabled = True  # output/no output switcher. Is controlled by -q/-qq cmd arg


class OptionalOutput(ContextDecorator):
    def __init__(self, active: bool):
        self.active = active

    def __enter__(self):
        global output_enabled
        if self.active:
            output_enabled = False

    def __exit__(self, exc_type, exc, exc_tb):
        if self.active:
            global output_enabled
            output_enabled = True


def blue(output: str) -> str:
    return Fore.BLUE + output + Style.RESET_ALL if colored_output else output


def green(output: str) -> str:
    return Fore.GREEN + output + Style.RESET_ALL if colored_output else output


def light_green(output: str) -> str:
    return Fore.LIGHTGREEN_EX + output + Style.RESET_ALL if colored_output else output


def red(output: str) -> str:
    return Fore.RED + output + Style.RESET_ALL if colored_output else output


def yellow(output: str) -> str:
    return Fore.YELLOW + output + Style.RESET_ALL if colored_output else output


def configure(log_level: str, use_color: bool = True):
    level = None
    if log_level == 'critical':
        level = logging.CRITICAL
    if log_level == 'error':
        level = logging.ERROR
    if log_level == 'warning':
        level = logging.WARNING
    if log_level == 'info':
        level = logging.INFO
    if log_level == 'debug':
        level = logging.DEBUG
    if not log_level:
        raise RuntimeError('Unknown logging level ' + log_level)
    logging.basicConfig(level=level)
    logging.getLogger("requests").setLevel(logging.WARNING)
    global colored_output
    colored_output = use_color


def get_logger() -> Logger:
    return logging.getLogger(catcher.APPNAME)


def debug(msg: str):
    if logging.DEBUG >= logging.root.level:
        log_storage.output('debug', msg)
    if output_enabled:
        get_logger().debug(_nested_output(msg))


def info(msg: str):
    if logging.INFO >= logging.root.level:
        log_storage.output('info', msg)
    if output_enabled:
        get_logger().info(_nested_output(msg))


def warning(msg: str):
    if logging.WARNING >= logging.root.level:
        log_storage.output('warning', msg)
    if output_enabled:
        get_logger().warning(_nested_output(msg))


def error(msg: str):
    if logging.ERROR >= logging.root.level:
        log_storage.output('error', msg)
    if output_enabled:
        get_logger().error(_nested_output(msg))


def critical(msg: str):
    if logging.CRITICAL >= logging.root.level:
        log_storage.output('critical', msg)
    if output_enabled:
        get_logger().critical(_nested_output(msg))


def _nested_output(msg):
    if log_storage.nesting_counter > 0:
        spaces = ''.join(['\t' for i in range(0, log_storage.nesting_counter - 1)])
        msg = spaces + '|--> ' + str(msg)
    return msg



import os
import stat
from os.path import join

from catcher.core.runner import Runner
from catcher.utils.module_utils import load_external_actions
from test.abs_test_class import TestClass


class ExternalTest(TestClass):
    def __init__(self, method_name):
        super().__init__('external', method_name)

    def test_run_external(self):
        self.write_module('math', '''#!/bin/bash
        one=$(echo ${1} | jq -r '.add.the')
        two=$(echo ${1} | jq -r '.add.to')
        echo $((${one} + ${two}))
        ''')
        self.populate_file('main.yaml', '''---
        steps:
            - math:
                add: {the: 1, to: 2}
                register: {sum: '{{ OUTPUT }}'}
            - check: {equals: {the: '{{ sum }}', is: 3}}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None, modules=[self.test_dir])
        self.assertTrue(runner.run_tests())

    def test_run_external_with_vars(self):
        self.write_module('math', '''#!/bin/bash
        one=$(echo ${1} | jq -r '.add.the')
        two=$(echo ${1} | jq -r '.add.to')
        echo $((${one} + ${two}))
        ''')
        self.populate_file('main.yaml', '''---
        variables:
            one: 1
            two: 2
        steps:
            - math:
                add: {the: '{{ one }}', to: '{{ two }}'}
                register: {sum: '{{ OUTPUT }}'}
            - check: {equals: {the: '{{ sum }}', is: 3}}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None, modules=[self.test_dir])
        self.assertTrue(runner.run_tests())

    def test_python_module(self):
        self.write_module('hello.py', '''from catcher.steps.external_step import ExternalStep
from catcher.steps.step import update_variables


class Hello(ExternalStep):
    """
    Very important and useful step. Says hello to input. Return as a string.
    Example usage.
    ::
        hello:
            say: 'John Doe'
            register: {greeting='{{ OUTPUT }}'}

    """
    @update_variables
    def action(self, includes: dict, variables: dict) -> (dict, str):
        body = self.simple_input(variables)
        person = body['say']
        return variables, 'hello {}'.format(person)
                ''')
        self.populate_file('main.yaml', '''---
                variables:
                    person: 'John Doe'
                steps:
                    - hello:
                        say: '{{ person }}'
                        register: {greeting: '{{ OUTPUT }}'}
                    - check: {equals: {the: '{{ greeting }}', is: 'hello John Doe'}}
                ''')
        load_external_actions(join(self.test_dir, 'hello.py'))
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None, modules=[self.test_dir])
        self.assertTrue(runner.run_tests())

    def test_run_jar(self):
        self.copy_resource('MyClass.jar')
        self.populate_file('main.yaml', '''---
                                        variables:
                                            person: 'John Doe'
                                        steps:
                                            - MyClass.jar:
                                                say: '{{ person }}'
                                                register: {greeting: '{{ OUTPUT }}'}
                                            - check: {equals: {the: '{{ greeting.strip() }}', is: 'hello John Doe'}}
                                        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None, modules=[self.test_resources])
        self.assertTrue(runner.run_tests())

    # make sure javac is available in the PATH to pass this test
    def test_compile_and_run_java(self):
        self.copy_resource('MyClass.java')
        self.copy_resource('OtherClass.java')
        self.populate_file('main.yaml', '''---
                                variables:
                                    person: 'John Doe'
                                steps:
                                    - MyClass.java:
                                        say: '{{ person }}'
                                        register: {greeting: '{{ OUTPUT }}'}
                                    - check: {equals: {the: '{{ greeting.strip() }}', is: 'hello John Doe'}}
                                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None, modules=[self.test_resources])
        self.assertTrue(runner.run_tests())

    # make sure kotlinc is available in the PATH to pass this test
    def test_compile_and_run_kotlin(self):
        self.copy_resource('hello.kt')
        self.populate_file('main.yaml', '''---
                                        variables:
                                            person: 'John Doe'
                                        steps:
                                            - hello.kt:
                                                say: '{{ person }}'
                                                register: {greeting: '{{ OUTPUT }}'}
                                            - check: {equals: {the: '{{ greeting.strip() }}', is: 'hello John Doe'}}
                                        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None, modules=[self.test_resources])
        self.assertTrue(runner.run_tests())

    def test_python_module_with_json(self):
        self.write_module('print.py', '''from catcher.steps.external_step import ExternalStep
from catcher.steps.step import update_variables

class Print(ExternalStep):

    @update_variables
    def action(self, includes: dict, variables: dict) -> (dict, str):
        body = self.simple_input(variables)
        data = body['the']
        return variables, 'data is: {}'.format(data)
                    ''')
        self.populate_file('inventory.yml', '''---
                complex_conf:
                  field: 'value'
                  json_field: '{"host": "http://minio:9000","aws_access_key_id":"minio","aws_secret_access_key":"minio123"}'
                ''')
        self.populate_file('main.yaml', '''---
                    steps:
                        - print:
                            the: '{{ complex_conf }}'
                    ''')
        load_external_actions(join(self.test_dir, 'print.py'))
        runner = Runner(self.test_dir,
                        join(self.test_dir, 'main.yaml'),
                        join(self.test_dir, 'inventory.yml'),
                        modules=[self.test_dir])
        self.assertTrue(runner.run_tests())

    def write_module(self, name: str, body: str):
        self.populate_file(name, body)
        path = join(self.test_dir, name)
        st = os.stat(path)
        os.chmod(path, st.st_mode | stat.S_IEXEC)



import json

from catcher.steps.step import Step, update_variables
from catcher.utils.misc import fill_template_str
from catcher.utils import external_utils


class External(Step):
    def __init__(self, _module: str = None, **kwargs) -> None:
        super().__init__(**kwargs)
        method = Step.filter_predefined_keys(kwargs)
        self.data = {method: kwargs[method]}
        self.module = _module

    @update_variables
    def action(self, includes: dict, variables: dict) -> tuple:
        """
        Call external script.

        :param includes: testcase's includes
        :param variables: variables
        :return: script's output
        """
        json_args = fill_template_str(json.dumps(self.data), variables)
        return variables, external_utils.run_cmd_simple(self.module, variables, args=[json_args])



from os.path import join

from catcher.core.runner import Runner
from test.abs_test_class import TestClass


class WaitTest(TestClass):
    def __init__(self, method_name):
        super().__init__('wait', method_name)

    def test_wait_fixed(self):
        self.populate_file('main.yaml', '''---
            steps:
                - echo: {from: '{{ NOW_TS }}', register: {before: '{{ OUTPUT }}'}}
                - wait: {seconds: 1}
                - echo: {from: '{{ NOW_TS }}', register: {after: '{{ OUTPUT }}'}}
                - check: {equals: {the: '{{ after - before >= 1 }}', is: true}}
            ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    def test_wait_service(self):
        """
        variables should be accessible both reads and writes for commands in for-block
        """
        self.populate_file('main.yaml', '''---
                    variables:
                        counter: 0
                    steps:
                        - wait: 
                            seconds: 1
                            for:
                                echo: {from: '{{ counter + 1}}', register: {counter: '{{ OUTPUT }}'}}
                        - check: {equals: {the: '{{ counter }}', is: 1}}
                    ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    def test_wait_success(self):
        """
        wait should return as soon as `for` succeeds
        """
        self.populate_file('main.yaml', '''---
                            variables:
                                counter: 0
                            steps:
                                - wait: 
                                    seconds: 1
                                    for:
                                        - echo: {from: '{{ counter + 1}}', register: {counter: '{{ OUTPUT }}'}}
                                        - check: {equals: {the: '{{ counter }}', is: 1}}
                                - check: {equals: {the: '{{ counter }}', is: 1}}
                            ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    def test_wait_failure(self):
        """
        wait should stop after delay if `for` fails
        """
        self.populate_file('main.yaml', '''---
                                    variables:
                                        counter: 0
                                    steps:
                                        - echo: {from: '{{ NOW_TS }}', register: {before: '{{ OUTPUT }}'}}
                                        - wait: 
                                            seconds: 1
                                            for:
                                                check: {equals: {the: '{{ counter }}', is: 1}}
                                    ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertFalse(runner.run_tests())



from os.path import join

import requests_mock
import test

from catcher.core.runner import Runner
from test.abs_test_class import TestClass


class ComposeTest(TestClass):
    def __init__(self, method_name):
        super().__init__('compose', method_name)

    @requests_mock.mock()
    def test_no_compose(self, m):
        m.get('http://test.com', status_code=500)
        self.populate_file('main.yaml', '''---
                    steps:
                        - http: {get: {url: 'http://test.com', response_code: 200}}
                    ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertFalse(runner.run_tests())

    def test_compose(self):
        self.populate_resource('docker-compose.yml', '''
        version: '3.1'
        services:
            mockserver:
                image: jamesdbloom/mockserver
                ports:
                - "8000:1080"
        ''')
        self.populate_file('main.yaml', '''---
                            steps:
                                 - wait:
                                    seconds: 5
                                    for:
                                        http:
                                            put:
                                                url: 'http://localhost:8000/mockserver/expectation'
                                                body:
                                                    httpRequest: {'path': '/some/path'}
                                                    httpResponse: {'body': 'hello world'}
                                                response_code: 201
                                - http: 
                                    get: 
                                        url: 'http://localhost:8000/mockserver/expectation',
                                        response_code: 200
                            ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None,
                        resources=join(test.get_test_dir(self.test_name), 'resources'))
        self.assertFalse(runner.run_tests())



from os.path import join

import requests_mock

from catcher.utils.file_utils import read_file

from catcher.core.runner import Runner
from test.abs_test_class import TestClass
from test.test_utils import check_file


class IncludeFilesTest(TestClass):
    def __init__(self, method_name):
        super().__init__('includes_files_test', method_name)

    # test variables from one include is available in another
    def test_vars_available_got(self):
        self.populate_file('main.yaml', '''---
        include: 
            - one.yaml
            - two.yaml
        ''')
        self.populate_file('one.yaml', '''---
        variables:
            foo: bar
        steps:
            - echo: {from: '{{ foo }}', to: one.output}
        ''')
        self.populate_file('two.yaml', '''---
        steps:
            - echo: {from: '{{ foo }}', to: two.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'one.output'), 'bar'))
        self.assertTrue(check_file(join(self.test_dir, 'two.output'), 'bar'))

    # test variables in include statement override file's and got variables
    def test_vars_override_manual(self):
        self.populate_file('main.yaml', '''---
        include: 
            - one.yaml
            - file: two.yaml
              variables:
                one: override1
                two: override2
        ''')
        self.populate_file('one.yaml', '''---
        variables:
            one: foo
        steps:
            - echo: {from: 'hello'}
        ''')
        self.populate_file('two.yaml', '''---
        variables:
            two: bar
        steps:
            - echo: 
                actions: 
                    - {from: '{{ one }}', to: one.output}
                    - {from: '{{ two }}', to: two.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'one.output'), 'override1'))
        self.assertTrue(check_file(join(self.test_dir, 'two.output'), 'override2'))

    # test variables in file override got variables
    def test_vars_override_file(self):
        self.populate_file('main.yaml', '''---
        include: 
            - one.yaml
            - two.yaml
        ''')
        self.populate_file('one.yaml', '''---
        variables:
            foo: bar
        steps:
            - echo: {from: 'hello'}
        ''')
        self.populate_file('two.yaml', '''---
        variables:
            foo: baz
        steps:
            - echo: {from: '{{ foo }}', to: one.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'one.output'), 'baz'))

    # test computed variables available in another
    def test_vars_available_computed(self):
        self.populate_file('main.yaml', '''---
        include: 
            - one.yaml
            - two.yaml
        ''')
        self.populate_file('one.yaml', '''---
        steps:
            - echo: {from: 'hello', register: {foo: 'bar'}}
            
        ''')
        self.populate_file('two.yaml', '''---
        steps:
            - echo: {from: '{{ foo }}', to: one.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'one.output'), 'bar'))

    # test variables in include statement override computed
    def test_vars_override_computed(self):
        self.populate_file('main.yaml', '''---
        include: 
            - one.yaml
            - file: two.yaml
              variables: 
                foo: baz
        ''')
        self.populate_file('one.yaml', '''---
        steps:
            - echo: {from: 'hello', register: {foo: 'bar'}}
            
        ''')
        self.populate_file('two.yaml', '''---
        steps:
            - echo: {from: '{{ foo }}', to: one.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'one.output'), 'baz'))

    # test if variables, pass to includes are filled in
    def test_vars_included_filled_in(self):
        self.populate_file('main.yaml', '''---
        include: 
            - file: one.yaml
              as: one
        steps:
            - echo: {from: '{{ RANDOM_STR }}', to: main.output, register: {uuid: '{{ OUTPUT }}'}}
            - run: 
                include: one
                variables: 
                    id: '{{ uuid }}'
        ''')
        self.populate_file('one.yaml', '''---
        steps:
            - echo: {from: '{{ id }}', to: one.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        main = read_file(join(self.test_dir, 'main.output'))
        one = read_file(join(self.test_dir, 'one.output'))
        self.assertEqual(main, one)
        self.assertTrue(one)

    def test_run_override_computed(self):
        self.populate_file('main.yaml', '''---
                include: 
                    - file: one.yaml
                      as: one
                steps:
                    - echo: {from: '{{ RANDOM_STR }}', to: main.output, register: {uuid: '{{ OUTPUT }}'}}
                    - run: 
                        include: one
                        variables: 
                            id: '{{ uuid }}'
                ''')
        self.populate_file('one.yaml', '''---
                variables:
                    id: 1234
                steps:
                    - echo: {from: '{{ id }}', to: one.output}
                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        main = read_file(join(self.test_dir, 'main.output'))
        one = read_file(join(self.test_dir, 'one.output'))
        self.assertEqual(main, one)
        self.assertTrue(one)

    def test_run_override_main(self):
        self.populate_file('main.yaml', '''---
                variables:
                    id: '{{ RANDOM_STR }}'
                include: 
                    - file: one.yaml
                      as: one
                steps:
                    - echo: {from: '{{ id }}', to: main.output}
                    - run: 
                        include: one
                ''')
        self.populate_file('one.yaml', '''---
                variables:
                    id: 1234
                steps:
                    - echo: {from: '{{ id }}', to: one.output}
                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        main = read_file(join(self.test_dir, 'main.output'))
        one = read_file(join(self.test_dir, 'one.output'))
        self.assertEqual(main, one)
        self.assertTrue(one)

    # all level of templates should be resolved when including variables
    def test_template_in_template(self):
        self.populate_file('main.yaml', '''---
                variables:
                    bar: '{{ RANDOM_STR }}'
                include: 
                    - file: one.yaml
                      as: one
                steps:
                    - echo: {from: '{{ bar }}', to: main.output}
                    - run: one
                ''')
        self.populate_file('one.yaml', '''---
                variables:
                    foo: '{{ bar }}'
                include: 
                    - file: two.yaml
                      as: two
                steps:
                    - echo: {from: '{{ foo }}', to: one.output}
                    - run: two
                ''')
        self.populate_file('two.yaml', '''---
                variables:
                    final: '{{ foo }}'
                steps:
                    - echo: {from: '{{ final }}', to: two.output}
                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        main = read_file(join(self.test_dir, 'main.output'))
        one = read_file(join(self.test_dir, 'one.output'))
        two = read_file(join(self.test_dir, 'two.output'))
        self.assertEqual(main, one)
        self.assertEqual(main, two)

    # all level of templates in resources should be resolved
    @requests_mock.mock()
    def test_template_in_template_resource(self, m):
        self.populate_file('main.yaml', '''---
                        variables:
                            bar: '{{ RANDOM_STR }}'
                        include: 
                            - file: one.yaml
                              as: one
                        steps:
                            - echo: {from: '{{ bar }}', to: main.output}
                            - run: one
                        ''')
        self.populate_file('one.yaml', '''---
                        variables:
                            foo: '{{ bar }}'
                        steps:
                            - http: 
                                post: 
                                    url: 'http://test.com'
                                    files:
                                        file: 'foo.json'
                        ''')
        self.populate_resource('foo.json', "{\"key\":\"{{ foo }}\"}")
        adapter = m.post('http://test.com')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())
        main = read_file(join(self.test_dir, 'main.output'))
        self.assertTrue("{\"key\":\"" + main + "\"}" in adapter.last_request.text)

    # predefined variable from include should be available in test already filled in,
    def test_variables_registered_in_include(self):
        self.populate_file('main.yaml', '''---
                include: 
                    - file: one.yaml
                      as: one
                steps:
                    - run: one
                    - echo: {from: '{{ generated_email }}', to: two.output}
                ''')

        self.populate_file('one.yaml', '''---
                variables:
                    generated_email: '{{ random("email") }}'
                steps:
                    - echo: {from: '{{ generated_email }}', to: one.output}
                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        two = read_file(join(self.test_dir, 'two.output'))
        one = read_file(join(self.test_dir, 'one.output'))
        self.assertEqual(two, one)
        self.assertNotEqual('{{ random("email") }}', one)



import subprocess
import sys
from os.path import isfile, join

from catcher.modules import Module
from catcher.utils.logger import debug


class Requirements(Module):
    """
    :Python requirements module:

    Will automatically install all packages from ``requirements.txt`` before running tests, if
    finds it in the root of the resource directory.

    """

    def __init__(self, resources_dir) -> None:
        super().__init__(resources_dir)

    def before(self, *args, **kwargs):
        requirements = self.find_resource_file()
        if requirements:
            debug('Installing requirements: {}'.format(requirements))
            subprocess.check_call([sys.executable, '-m', 'pip', 'install', '-r', requirements])

    def after(self, *args, **kwargs):
        # does nothing
        pass

    def filter_resource(self, file: str) -> bool:
        # use only default docker-compose.yml for now, will change it in future if needed.
        return isfile(join(self._resources, file)) and file == 'requirements.txt'



TIME_MAPPING = [
    ('days', 86400),
    ('hours', 3600),
    ('minutes', 60),
    ('seconds', 1),
    ('microseconds', 0.001),
    ('milliseconds', 0.000001),
    ('nanoseconds', 0.000000001)
]


def to_seconds(body: dict) -> int:
    return sum([compute_time(t, body, m) for (t, m) in TIME_MAPPING])


def compute_time(key: str, body: dict, to_second: float) -> int:
    return body.get(key, 0) * to_second



from contextlib import ContextDecorator
from os.path import join
from shutil import copyfile

import test
from catcher.core.runner import Runner
from test.abs_test_class import TestClass
from catcher.utils import module_utils
from catcher.utils import file_utils


class grpc_server(ContextDecorator):
    def __init__(self, server) -> None:
        super().__init__()
        self.server = server

    def __enter__(self):
        resources = join(test.get_test_dir('grpc'), 'private_resource')
        mod = module_utils.load_external_actions(join(resources, self.server))
        self.server = mod.server
        self.server.start()
        return self

    def __exit__(self, *exc):
        self.server.stop(0)
        return False


class GRPCTest(TestClass):
    def __init__(self, method_name):
        super().__init__('grpc', method_name)

    def setUp(self):
        super().setUp()
        file_utils.ensure_empty(join(test.get_test_dir(self.test_name), 'private_resource'))
        self.copy_resource('calculator.py')
        self.copy_resource('calculator.proto')
        self.copy_resource('greeter.py')
        self.copy_resource('greeter.proto')
        from grpc.tools import command
        command.build_package_protos(join(test.get_test_dir(self.test_name), 'private_resource'), strict_mode=True)

    def copy_resource(self, resource_name):
        # copy in private non-resource dir instead
        copyfile(join(self.global_resource_dir, resource_name),
                 join(join(test.get_test_dir(self.test_name), 'private_resource'), resource_name))

    @grpc_server('calculator.py')
    def test_call_simple(self):
        TestClass.copy_resource(self, 'calculator.proto')
        self.populate_file('main.yaml', '''---
                    variables:
                        my_value: 4
                    steps:
                        - wait:
                            seconds: 5
                            for:
                                - grpc:
                                    call:
                                        url: 'localhost:50051'
                                        function: calculator.squareroot
                                        schema: 'calculator.proto'
                                        data: {'value': '{{ my_value }}'}
                                    register: {value: '{{ OUTPUT.value }}'}
                        - check: 
                            equals: {'the': '{{ value }}', 'is': 2.0}
                    ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    @grpc_server('greeter.py')
    def test_call_complex(self):
        TestClass.copy_resource(self, 'greeter.proto')
        self.populate_file('main.yaml', '''---
                            steps:
                                - wait:
                                    seconds: 5
                                    for:
                                        - grpc:
                                            call:
                                                url: 'localhost:50051'
                                                function: greeter.greet
                                                schema: 'greeter.proto'
                                                data: 
                                                    result:
                                                        url: '{{ my_url }}'
                                                        title: 'test'
                                                        snippets: 'test2'
                                            register: {value: '{{ OUTPUT.name }}'}
                                - check: 
                                    equals: {'the': '{{ value }}', 'is': 'Result for test contains test2'}
                            ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    @grpc_server('greeter.py')
    def test_call_variables(self):
        TestClass.copy_resource(self, 'greeter.proto')
        self.populate_file('main.yaml', '''---
                            variables:
                                data:
                                    result: 
                                        url: 'my'
                                        title: 'test'
                                        snippets: 'test2'
                            steps:
                                - wait:
                                    seconds: 5
                                    for:
                                        - grpc:
                                            call:
                                                url: 'localhost:50051'
                                                function: greeter.greet
                                                schema: 'greeter.proto'
                                                data: '{{ data }}'
                                            register: {value: '{{ OUTPUT.name }}'}
                                - check: 
                                    equals: {'the': '{{ value }}', 'is': 'Result for test contains test2'}
                            ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())



from abc import abstractmethod

from catcher.steps.step import Step, update_variables, SERVICE_KEYS
from catcher.utils.logger import debug
from catcher.utils.misc import fill_template, fill_template_str
from catcher.utils.module_utils import get_all_subclasses_of


class Operator(object):
    def __init__(self, body: dict, negative=False) -> None:
        self.subject = body
        self.negative = negative

    @property
    def body(self) -> any:
        return self.__class__.__name__.lower()

    @abstractmethod
    def operation(self, variables: dict) -> bool:
        pass

    @staticmethod
    def find_operator(source: dict or str) -> 'Operator':
        if isinstance(source, str):
            operator_str = 'equals'
        else:
            [operator_str] = source.keys()
        operators = get_all_subclasses_of(Operator)
        named = dict([(o.__name__.lower(), o) for o in operators])
        if operator_str not in named:
            raise RuntimeError('No ' + operator_str + ' available')
        cls = named[operator_str]
        return cls(source)


class Equals(Operator):
    """
    Fail if elements are not equal

    :Input:
    :the: value
    :is: variable to compare
    :is_not: inverted `is`. Only one can be used at a time.

    :Examples:

    Check 'bar' equals variable 'foo'
    ::

        check: {equals: {the: 'bar', is: '{{ foo }}'}}

    Check list's third element is not greater than 2.
    ::

        check: {equals: {the: '{{ list[2] > 2 }}', is_not: true}}

    """

    def operation(self, variables: dict) -> bool:
        if isinstance(self.subject, str):
            subject = fill_template(self.subject, variables)
            source = True
        else:
            body = self.subject[self.body]
            if isinstance(body, str):
                body = Equals.to_long_form(body, True)
            subject = fill_template(body['the'], variables)
            source = fill_template(self.determine_source(body), variables)
        result = source == subject
        if self.negative:
            result = not result
        if not result:
            debug(str(source) + ' is not equal to ' + str(subject))
        return result

    def determine_source(self, body: dict):
        if 'is' in body:
            return body['is']
        self.negative = True
        return body['is_not']

    @staticmethod
    def to_long_form(source: any, value: any):
        return {'the': source, 'is': value}


class Contains(Operator):
    """
    Fail if list of dictionary doesn't contain the value

    :Input:
    :the: value to contain
    :in: variable to check
    :not_in: inverted `in`. Only one can be used at a time.

    :Examples:

    Check 'a' not in variable 'list'
    ::

        check:
            contains: {the: 'a', not_in: '{{ list }}'}

    Check variable 'dict' has key `a`.
    ::

        check:
            contains: {the: 'a', in: '{{ dict }}'}

    """

    def operation(self, variables: dict):
        body = self.subject[self.body]
        source = fill_template(self.determine_source(body), variables)
        subject = fill_template(body['the'], variables)
        if isinstance(source, str):
            result = str(subject) in source
        else:
            result = subject in source
        if self.negative:
            result = not result
        if not result:
            debug(str(subject) + ' is not in ' + str(source))
        return result

    def determine_source(self, body: dict):
        if 'in' in body:
            return body['in']
        self.negative = True
        return body['not_in']

    @staticmethod
    def to_long_form(source: any, value: any):
        return {'the': source, 'in': value}


class And(Operator):
    """
    Fail if any of the conditions fails.

    :Input: The list of other checks.

    :Examples:

    This is the same as `1 in list and list[1] != 'b' and list[2] > 2`
    ::

        check:
            and:
                - contains: {the: 1, in: '{{ list }}'}
                - equals: {the: '{{ list[1] }}', is_not: 'b'}
                - equals: {the: '{{ list[2] > 2 }}', is_not: true}

    """

    @property
    def end(self) -> bool:
        return False

    def operation(self, variables) -> bool:
        operators = self.subject[self.body]  # or or and
        for operator in operators:
            next_operation = Operator.find_operator(operator)
            if next_operation.operation(variables) == self.end:
                return self.end
        return True


class Or(And):
    """
    Fail if all conditions fail.

    :Input: The list of other checks.

    :Examples:

    This is the same as `1 in list or list[1] != 'b' or list[2] > 2`
    ::

        check:
            or:
                - contains: {the: 1, in: '{{ list }}'}
                - equals: {the: '{{ list[1] }}', is_not: 'b'}
                - equals: {the: '{{ list[2] > 2 }}', is_not: true}

    """

    @property
    def end(self):
        return True


class All(Operator):
    """
    Fail if any check on the iterable fail.

    :Input:
    :of: The source to check. Can be list or dictionary.
    :<check>: Check to perform on each element of the iterable.

    :Examples:

    Pass if all elements of `var` has `k` == `a`
    ::

        check:
            all:
                of: '{{ var }}'
                equals: {the: '{{ ITEM.k }}', is: 'a'}

    """

    def operator(self, data):
        return all(data)

    def operation(self, variables) -> bool:
        body = self.subject[self.body]
        source = fill_template(body['of'], variables)
        if isinstance(source, list):
            elements = source
        elif isinstance(source, dict):
            elements = source.items()
        else:
            debug(str(source) + ' not iterable')
            return False
        results = []
        for element in elements:
            oper_body = dict([(k, v) for (k, v) in body.items() if k != 'of'])
            [next_operator] = oper_body.keys()
            if not isinstance(oper_body[next_operator], dict):  # terminator in short form
                if next_operator == 'equals':
                    oper_body[next_operator] = Equals.to_long_form('{{ ITEM }}', oper_body[next_operator])
                if next_operator == 'contains':
                    oper_body[next_operator] = Contains.to_long_form('{{ ITEM }}', oper_body[next_operator])
            next_operation = Operator.find_operator(oper_body)
            variables['ITEM'] = element
            results.append(next_operation.operation(variables))
        return self.operator(results)


class Any(All):
    """
    Fail if all checks on the iterable fail.

    :Input:
    :of: The source to check. Can be list or dictionary.
    :<check>: Check to perform on each element of the iterable.

    :Examples:

    Fail if `var` doesn't contain element with `k` == `a`
    ::

        check:
            any:
                of: '{{ var }}'
                equals: {the: '{{ ITEM.k }}', is: 'a'}

    """

    def operator(self, data):
        return any(data)


class Check(Step):
    """
    Run check and fail if it was not successful.

    There are two types of checks: terminators and nodes. Terminators like `equals` or `contains`
    just perform checks while nodes contain like `all`, `any`, `or` and others contain other checks.

    Check has a short form
    ::

        check: '{{ variable }}'

    which equals
    ::

        check:
            equals: {the: '{{ variable }}', is: true}

    """

    def __init__(self, _body=None, **kwargs: dict) -> None:
        super().__init__(**kwargs)
        if _body:
            self.subject = _body
        else:
            [subject] = [{k: v} for k, v in kwargs.items() if k not in SERVICE_KEYS and not k.startswith('_')]
            self.subject = subject

    @update_variables
    def action(self, includes: dict, variables: dict) -> dict:
        operator = Operator.find_operator(self.subject)
        if not operator.operation(variables):
            raise RuntimeError('operation ' + fill_template_str(self.subject, variables) + ' failed')
        return variables



import itertools
import time
from concurrent.futures import ThreadPoolExecutor
from copy import deepcopy
from time import sleep
from catcher.steps.step import Step, update_variables
from catcher.utils.time_utils import to_seconds
from catcher.utils.logger import debug


class Wait(Step):
    """
    Wait for a static delay or until some substep finished successfully. Is extremely useful for testing asnyc systems
    or when you are waiting for some service to launch.

    :Input:

    :days: several days
    :hours: several hours
    :minutes: several minutes
    :seconds: several seconds
    :microseconds: several microseconds
    :milliseconds: several milliseconds
    :nanoseconds: several nanoseconds
    :for: (list of actions) will repeat them till they all finishes successfully. Will fail if time ends.

    :Examples:

    Wait for 1 minute 30 seconds
    ::

        wait: {minutes: 1, seconds: 30}

    Wait for http to be ready. Will repeat inner **http** step till is succeeded or fails after 5 seconds
    ::

        wait:
            seconds: 5
            for:
                http:
                    put:
                        url: 'http://localhost:8000/mockserver/expectation'
                        body:
                            httpRequest: {'path': '/some/path'}
                            httpResponse: {'body': 'hello world'}
                        response_code: 201

    Wait for postgres to be populated
    ::

        wait:
            seconds: 30
            for:
                - postgres:
                      request:
                          conf: '{{ pg_conf }}'
                          query: 'select count(*) from users'
                      register: {documents: '{{ OUTPUT }}'}
                - check: {equals: {the: '{{ documents }}', is_not: 0}}
    """

    def __init__(self, _get_action=None, _get_actions=None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.delay = to_seconds(kwargs)
        self._actions = None
        if 'for' in kwargs:
            wait_for = kwargs['for']
            if len(wait_for) == 1:
                if isinstance(wait_for, list):
                    wait_for = wait_for[0]
                [loop_action] = wait_for.keys()
                self._actions = [_get_action((loop_action, wait_for[loop_action]))]
            elif len(wait_for) > 1:
                self._actions = list(itertools.chain.from_iterable([_get_actions(act) for act in wait_for]))

    @update_variables
    def action(self, includes: dict, variables: dict) -> dict:
        if self._actions is not None:
            # run thread and either exit on success or wait fixed time
            pool = ThreadPoolExecutor(max_workers=1)
            future = pool.submit(self.run_loop, includes, variables)
            try:
                return future.result()
            except TimeoutError:
                return variables
        else:  # wait fixed time
            sleep(self.delay)
            return variables

    def run_loop(self, includes, variables):
        output = variables
        repeat = True
        start = time.time()
        while repeat:
            if time.time() > start + self.delay:  # time limit reached
                raise Exception('Time limit reach with no success from substeps')
            loop_vars = deepcopy(output)  # start every loop from the same variables
            try:
                for action in self._actions:
                    loop_vars = action.action(includes, loop_vars)
                repeat = False
                output = loop_vars  # if loop was successful - return modified variables
            except Exception as e:
                debug('Wait step failure {}'.format(e))
        return output



from os.path import join

from catcher.core.runner import Runner
from test.abs_test_class import TestClass


class StopTest(TestClass):
    def __init__(self, method_name):
        super().__init__('stop', method_name)

    def test_stop(self):
        self.populate_file('main.yaml', '''---
            variables:
                counter: 0
            steps:
                - echo: {from: '{{ counter + 1 }}', register: {counter: '{{ OUTPUT }}'}}
                - stop: 
                    if: 
                        equals: '{{ counter > 1 }}'
                - echo: {from: '{{ counter + 1 }}', register: {counter: '{{ OUTPUT }}'}}
                - stop: 
                    if: 
                        equals: '{{ counter > 1 }}'
                - echo: {from: '{{ counter + 1 }}', register: {counter: '{{ OUTPUT }}'}}
                - check: {equals: {the: '{{ counter }}', is: 2}}
            ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    def test_stop_from_include(self):
        self.populate_file('main.yaml', '''---
        include: 
            file: migration.yaml
            as: migrate
        variables:
            counter: 0
        steps:
            - run: migrate
            - check: {equals: {the: '{{ counter }}', is: 2}}
            - run: migrate
            - check: {equals: {the: '{{ counter }}', is: 2}}
        ''')
        self.populate_file('migration.yaml', '''---
        steps:
            - echo: {from: '{{ counter + 1 }}', register: {counter: '{{ OUTPUT }}'}}
            - stop: 
                if: 
                    equals: '{{ counter > 1 }}'
            - echo: {from: '{{ counter + 1 }}', register: {counter: '{{ OUTPUT }}'}}
            - stop: 
                if: 
                    equals: '{{ counter > 1 }}'
            - echo: {from: '{{ counter + 1 }}', register: {counter: '{{ OUTPUT }}'}}
            - check: {equals: {the: '{{ counter }}', is: 2}}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())



import time
import traceback
from typing import Union

from catcher.steps.step import Step, SkipException
from catcher.steps.stop import StopException
from catcher.utils import logger
from catcher.utils.logger import debug, info
from catcher.utils.misc import fill_template_str
from catcher.steps.check import Operator
from catcher.core.step_factory import StepFactory


class Test:
    """
    Testcase. Contains variables, includes, steps and final actions.

    :ignore: if true will ignore this test. A condition based on :meth:`catcher.steps.check.Check` can be used to
     compute ignore condition dynamically.

    :include: other test to run. See :meth:`catcher.core.include.Include`. Can be a string in case of single
      include or a list of include in case of multiple. Each of them will be run before the test passing variables to
      each other and, finally, to the test. In case of `run_on_include` include's property is false or `as` alias is set
      include won't be run before the test. You can run such include via :meth:`catcher.steps.run.Run` step later.
      See :doc:`includes` for more info.

    :variables: test local variables which override inventory variables. These variables will be available only in this
     test or in test which includes this test. Variables itself can contain templates.
     See :doc:`variables` for more info.

    :steps: A list of test actions which will be run one by one. They can use variables and support templates. Each
     step can register it's output as a new variable. See :meth:`catcher.steps.step.Step` for available options.
     See :doc:`internal_modules` and `external modules <https://catcher-modules.readthedocs.io/en/latest/index.html>`_
     for more info.

    :finally: A list of clean up actions which will be run after test finishes execution. Condition for every clean up
     action run can be set (by default they run in any case).

    See :doc:`tests` for more info and examples.

    """

    def __init__(self,
                 path: str,
                 file: str,
                 includes: dict = None,
                 variables: dict = None,
                 config: dict = None,
                 steps: list = None,
                 final: list = None,
                 ignore: Union[bool, str, dict] = None) -> None:
        self.path = path
        self.final = final
        self.file = file
        self.includes = includes
        self.config = config
        self.variables = variables
        self.steps = steps
        self.ignore = ignore
        self.include = None  # if this test is include it will refer to `Include` class

    def check_ignored(self):
        if self.ignore:
            if not isinstance(self.ignore, bool):
                self.ignore = Operator.find_operator(self.ignore).operation(self.variables)
            if self.ignore:
                raise SkipException('Test ignored')

    def run(self, tag=None, raise_stop=False) -> dict:
        for step in self.steps:
            if not self._run_step(step, tag, raise_stop):
                break
        return self.variables

    def run_finally(self, result: bool):
        for step in self.final:
            if not self._run_step(step, result=result):
                return

    def _run_step(self, step, tag=None, raise_stop=False, result=None) -> bool:
        [action] = step.keys()
        ignore_errors = get_or_default('ignore_errors', step[action], False)
        if tag is not None:  # skip if tag specified
            step_tag = get_or_default('tag', step[action], None)
            if step_tag != tag:
                return True
        if result is not None:  # skip if result is specified (step is final)
            run_type = get_or_default('run_if', step[action], 'always')
            if (not result and run_type == 'pass') or (result and run_type == 'fail'):
                debug('Skip final action')
                return True
        actions = StepFactory().get_actions(self.path, step)
        for action_object in actions:
            # override all variables with cmd variables
            if not self._run_actions(step, action, action_object, self.variables, raise_stop, ignore_errors):
                return False
        return True

    def _run_actions(self, step, action, action_object, variables, raise_stop, ignore_errors) -> bool:
        action_name = get_action_name(action, action_object, variables)
        start = time.process_time()
        try:
            logger.log_storage.new_step(step, variables)
            action_object.check_skip(variables)
            self.variables = action_object.action(self.includes, variables)
            # repeat for run (variables were computed after name)
            action_name = get_action_name(action, action_object, self.variables)
            info('Step ' + action_name + get_timing(start) + logger.green(' OK'))
            logger.log_storage.step_end(step, self.variables)
            return True
        except StopException as e:  # stop a test without error
            if raise_stop:  # or raise error if configured
                logger.log_storage.step_end(step, variables, success=False, output=str(e))
                raise e
            debug('Skip ' + action_name + ' due to ' + str(e))
            info('Step ' + action_name + get_timing(start) + logger.green(' OK'))
            logger.log_storage.step_end(step, self.variables, success=True, output=str(e))
            return False  # stop current test
        except SkipException as e:  # skip this step
            info('Step ' + action_name + logger.yellow(' skipped'))
            logger.log_storage.step_end(step, self.variables, success=True, output=str(e))
            return True
        except Exception as e:
            if ignore_errors:  # continue actions & steps execution
                debug('Step ' + action_name + get_timing(start) + logger.red(' failed') + ', but we ignore it')
                logger.log_storage.step_end(step, variables, success=True)
                return True
            else:
                info('Step ' + action_name + get_timing(start) + logger.red(' failed: ') + str(e))
                debug(traceback.format_exc())
                logger.log_storage.step_end(step, variables, success=False, output=str(e))
                raise e

    def __repr__(self) -> str:
        return str(self.steps)


def get_or_default(key: str, body: dict or str, default: any) -> any:
    if isinstance(body, dict):
        return body.get(key, default)
    else:
        return default


def get_action_name(action_type: str, action: Step, variables: dict):
    if action.name is not None:
        return fill_template_str(action.name, variables)
    return action_type


def get_timing(start_time) -> str:
    delta = time.process_time() - start_time
    if delta > 60:
        return ' [{:.0f}m {:.0f}s]'.format(delta / 60, delta - 60 * int(delta / 60))
    return ' [{:.0f}s]'.format(delta)



from setuptools import setup, find_packages

import catcher


def get_requirements() -> list:
    with open('requirements.txt', 'r') as f:
        return f.readlines()


setup(name=catcher.APPNAME,
      version=catcher.APPVSN,
      description='Microservices automated test tool.',
      author=catcher.APPAUTHOR,
      author_email='valerii.tikhonov@gmail.com',
      url='https://github.com/comtihon/catcher',
      packages=find_packages(),
      install_requires=get_requirements(),
      include_package_data=True,
      package_data={'catcher': ['resources/*']},
      entry_points={
          'console_scripts': [
              'catcher=catcher.__main__:main'
          ]},
      classifiers=[
          'Intended Audience :: Developers',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Topic :: Software Development :: Testing'
      ],
      extras={'compose': ["docker-compose==1.24.*"]}
      )



import ast
import datetime
import json
import random
import time
import uuid
# noinspection PyUnresolvedReferences
from uuid import UUID  # for UUID as object parsing
from collections.abc import Iterable
from types import ModuleType
from typing import Union

from jinja2 import Template, UndefinedError

from catcher.utils import module_utils
from catcher.utils.logger import debug
from catcher.core.filters_factory import FiltersFactory


def merge_two_dicts(x, y):
    if not x:
        return y
    if not y:
        return x
    return {**x, **y}


def report_override(variables: dict, override: dict):
    existing = set(variables)
    replace = set(override)
    return list(existing.intersection(replace))


def try_get_objects(source: str or dict or list):
    got = try_get_object(source)  # "'[1,2,3]'" -> '[1,2,3]' -> [1,2,3]
    got = try_get_object(got)  # '[1,2,3]' -> [1,2,3]
    if isinstance(got, dict):
        return dict([(k, try_get_objects(v)) for k, v in got.items()])
    if isinstance(got, list):
        return [try_get_objects(v) for v in got]
    return got


def try_get_object(source: str or dict or list):
    if isinstance(source, str):
        try:  # try python term '{key: "value"}'
            evaled = eval_datetime(source)
            if isinstance(evaled, ModuleType) or callable(evaled):  # for standalone 'string' var or 'id' bif
                return source
            source = evaled
        except Exception:
            try:  # try json object '{"key" : "value"}'
                source = json.loads(source)
            except ValueError:
                return source
    return source


def fill_template_recursive(source: Union[dict, list, str], variables: dict, glob=None, globs_added=None) \
        -> Union[dict, list, str]:
    if isinstance(source, dict):
        return dict([(fill_template_recursive(k, variables, glob, globs_added),
                      fill_template_recursive(v, variables, glob, globs_added)) for k, v in source.items()])
    if isinstance(source, list):
        return [fill_template_recursive(v, variables, glob, globs_added) for v in source]
    return fill_template(source, variables, glob, globs_added)


def fill_template(source: str, variables: dict, isjson=False, glob=None, globs_added=None) -> str:
    if not globs_added:
        globs_added = set()
    if isinstance(source, str):
        source = render(source, inject_builtins(variables))
        if isjson:  # do not parse json string back to objects
            return source
        try:
            evaled = format_datetime(eval_datetime(source, glob))
            if not isinstance(evaled, ModuleType) and not callable(evaled):  # for standalone 'string' var or 'id' bif
                source = evaled
        except NameError as e:  # try to import missing package and rerun this code
            if 'is not defined' in str(e):
                name = str(e).split("'")[1]
                if name not in globs_added:
                    # f.e. tzinfo=psycopg2.tz.FixedOffsetTimezone for datetime
                    glob = module_utils.add_package_to_globals(name, glob, warn_missing_package=False)
                    globs_added.add(name)
                    filled = fill_template(source, variables, isjson, glob=glob, globs_added=globs_added)
                    if not isinstance(filled, ModuleType) and not callable(filled):
                        return filled  # for standalone 'string' var or 'id' bif
        except Exception:
            pass
    return source


def fill_template_str(source: any, variables: dict) -> str:
    rendered = render(str(source), inject_builtins(variables))
    if rendered != source:
        return fill_template_str(rendered, variables)
    return rendered


def eval_datetime(astr, glob=None):
    if glob is None:
        glob = globals()
    try:
        tree = ast.parse(astr)
    except SyntaxError:
        raise ValueError(astr)
    for node in ast.walk(tree):
        if isinstance(node, (ast.Module, ast.Expr, ast.Dict, ast.Str,
                             ast.Attribute, ast.Num, ast.Name, ast.Load, ast.Tuple)): continue
        if (isinstance(node, ast.Call)
                and isinstance(node.func, ast.Attribute)
                and node.func.attr == 'datetime'): continue
        pass
    return eval(astr, glob)


def format_datetime(iterable):
    if not isinstance(iterable, Iterable) or isinstance(iterable, str):
        if isinstance(iterable, datetime.datetime):
            return iterable.strftime('%Y-%m-%d %H:%M:%S.%f')
        return iterable
    else:
        if isinstance(iterable, dict):
            return dict([(format_datetime(k), format_datetime(v)) for k, v in iterable.items()])
        elif isinstance(iterable, tuple):
            return tuple([format_datetime(i) for i in iterable])
        return [format_datetime(i) for i in iterable]


def inject_builtins(variables: dict) -> dict:
    variables_copy = dict(variables)
    variables_copy['RANDOM_STR'] = str(uuid.uuid4())
    variables_copy['RANDOM_INT'] = random.randint(-2147483648, 2147483648)
    ts = round(time.time(), 6)  # from timestamp uses rounding, so we should also use it here, to make them compatible
    variables_copy['NOW_TS'] = ts
    variables_copy['NOW_DT'] = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%dT%H:%M:%S0+0000')
    return variables_copy


def render(source: str, variables: dict) -> str:
    template = Template(source)
    holder = FiltersFactory()
    for filter_mod, value in holder.filters.items():
        template.environment.filters[filter_mod] = value
    for fun_mod, value in holder.functions.items():
        template.globals[fun_mod] = value
    try:
        return template.render(variables)
    except UndefinedError as e:
        debug(e.message)
        return source



import os
from copy import deepcopy
from typing import Optional

from catcher.utils.logger import debug
from catcher.utils.misc import try_get_object, fill_template_str, report_override, merge_two_dicts


class VariablesHolder:
    """
    Variables override priority:
    1. cmd_env is passed as command line arguments
    2. newly registered variables
    3. test-local variables
    4. inventory variables
    5. environment variables
    """

    def __init__(self,
                 path: str,
                 system_environment=None,
                 inventory_vars: dict = None,
                 cmd_env: Optional[dict] = None,
                 resources: Optional[dict] = None) -> None:
        system_vars = system_environment or {}
        if system_vars:
            debug('Use system variables: ' + str(list(system_vars.keys())))
            self._variables = system_vars
        else:
            self._variables = {}
        inventory = try_get_object(fill_template_str(inventory_vars, self._variables))  # fill env vars
        self._cmd_env = cmd_env or {}
        self._variables.update(inventory)
        self._variables['CURRENT_DIR'] = path
        self._variables['RESOURCES_DIR'] = resources or os.path.join(path, 'resources')

    @property
    def variables(self) -> dict:
        """
        deep copy of inventory & environment variables
        """
        return deepcopy(self._variables)

    def prepare_variables(self, test, global_variables: dict):
        """
        Create variables for test
        :param global_variables:
        :param test: test with test variables
        """
        # system env + inventory
        # (test template is filled in based on system, inventory & cmd vars)
        test.variables = try_get_object(fill_template_str(test.variables,
                                                          merge_two_dicts(global_variables, self._cmd_env)))
        # test local
        global_variables.update(test.variables)
        # include vars override test local
        global_variables.update(self._prepare_include_vars(test, global_variables))
        # cmd_env override everything
        global_variables.update(self._cmd_env)
        test.variables = global_variables

    @staticmethod
    def _prepare_include_vars(test, global_variables):
        if test.include:
            include_vars = try_get_object(fill_template_str(test.include.variables, global_variables))
        else:
            include_vars = {}
        override_keys = report_override(test.variables, include_vars)
        if override_keys:
            debug('Include variables override these variables: ' + str(override_keys))
        return include_vars



import os
from os.path import join

from catcher.core.runner import Runner
from test.abs_test_class import TestClass
from test.test_utils import check_file


class IncludeFilesTest(TestClass):
    def __init__(self, method_name):
        super().__init__('includes_files_test', method_name)

    # test include one file, run on include
    def test_simple_include(self):
        self.populate_file('main.yaml', '''---
        include: simple_file.yaml
        ''')
        self.populate_file('simple_file.yaml', '''---
        variables:
            foo: bar
        steps:
            - echo: {from: '{{ foo }}', to: foo.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'foo.output'), 'bar'))

    # test include one file, no run on include
    def test_dict_include(self):
        self.populate_file('main.yaml', '''---
        include: 
            file: simple_file.yaml
            run_on_include: false
        ''')
        self.populate_file('simple_file.yaml', '''---
        variables:
            foo: bar
        steps:
            - echo: {from: '{{ foo }}', to: foo.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(not os.path.exists(join(self.test_dir, 'foo.output')))

    # test include multiple files
    def test_multiple_includes(self):
        self.populate_file('main.yaml', '''---
        include: 
            - file: simple_file.yaml
              run_on_include: false
            - other_simple_file.yaml
        ''')
        self.populate_file('simple_file.yaml', '''---
        variables:
            foo: bar
        steps:
            - echo: {from: '{{ foo }}', to: foo.output}
        ''')
        self.populate_file('other_simple_file.yaml', '''---
        variables:
            foo: baz
        steps:
            - echo: {from: '{{ foo }}', to: other.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(not os.path.exists(join(self.test_dir, 'foo.output')))
        self.assertTrue(check_file(join(self.test_dir, 'other.output'), 'baz'))

    # test includes with includes
    def test_recursive_includes(self):
        self.populate_file('main.yaml', '''---
        include: 
            - simple_file.yaml
        ''')
        self.populate_file('simple_file.yaml', '''---
        include: 
            - other_simple_file.yaml
        variables:
            foo: bar
        steps:
            - echo: {from: '{{ foo }}', to: foo.output}
        ''')
        self.populate_file('other_simple_file.yaml', '''---
        variables:
            foo: baz
        steps:
            - echo: {from: '{{ foo }}', to: other.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'foo.output'), 'bar'))
        self.assertTrue(check_file(join(self.test_dir, 'other.output'), 'baz'))

    # test fail on circular includes
    def test_circular_includes(self):
        self.populate_file('main.yaml', '''---
        include: 
            - simple_file.yaml
        ''')
        self.populate_file('simple_file.yaml', '''---
        include: 
            - other_simple_file.yaml
        variables:
            foo: bar
        steps:
            - echo: {from: '{{ foo }}', to: foo.output}
        ''')
        self.populate_file('other_simple_file.yaml', '''---
        include: 
            - simple_file.yaml
        variables:
            foo: baz
        steps:
            - echo: {from: '{{ foo }}', to: other.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertFalse(runner.run_tests())

    # includes from multiple places shouldn't throw curcular dependencies error
    def test_include_multiple_places(self):
        """
        main |--> include1 --> include2
             |--> include2
        """
        self.populate_file('main.yaml', '''---
                include: 
                        - file: include1.yml
                          as: inc1
                        - file: include2.yml
                          as: inc2
                ''')
        self.populate_file('include1.yml', '''---
                include:
                    - file: include2.yml
                      as: inc2
        ''')
        self.populate_file('include2.yml', '''---
        variables:
            foo: 123
        steps:
            - echo: {from: '{{ foo }}', to: other.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    # test run on include and later by alias
    def test_run_on_include_and_later(self):
        self.populate_file('main.yaml', '''---
        include: 
            file: simple_file.yaml
            variables:
                file: include
            as: test
            run_on_include: true
        steps:
            - run: 
                include: test
                variables: 
                    file: other
        ''')
        self.populate_file('simple_file.yaml', '''---
        steps:
            - echo: {from: 'hello', to: '{{ file }}.output'}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'include.output'), 'hello'))
        self.assertTrue(check_file(join(self.test_dir, 'other.output'), 'hello'))

    def test_include_skipped_test(self):
        self.populate_file('main.yaml', '''---
                include: simple_file.yaml
                steps:
                    - echo: {from: '12', to: bar.output}
                ''')
        self.populate_file('simple_file.yaml', '''---
                ignore: true
                variables:
                    foo: bar
                steps:
                    - echo: {from: '{{ foo }}', to: foo.output}
                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertFalse(os.path.exists(join(self.test_dir, 'foo.output')))
        self.assertTrue(check_file(join(self.test_dir, 'bar.output'), '12'))



from catcher.steps.check import Operator

from catcher.steps.step import Step, update_variables


class StopException(Exception):
    def __init__(self, message):
        super().__init__(message)


class Stop(Step):
    """
    Stop a test without error

    :Input:

    :if: condition

    :Examples:

    Stop execution if migration was applied.
    ::

        steps:
            - postgres:
                request:
                    conf: '{{ migrations_postgres }}'
                    query: "select count(*) from migration where hash = '{{ TEST_NAME }}';"
                register: {result: '{{ OUTPUT }}'}
                tag: check
                name: 'check_migration_{{ TEST_NAME }}'
            - stop:
                if:
                    equals: {the: '{{ result }}', is: 1}
            - postgres:
                request:
                    conf: '{{ migrations_postgres }}'
                    query: "insert into migration(id, hash) values(1, '{{ TEST_NAME }}');"
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.end_if = kwargs['if']

    @update_variables
    def action(self, includes: dict, variables: dict) -> dict:
        operator = Operator.find_operator(self.end_if)
        if operator.operation(variables):
            raise StopException(str(self.end_if) + ' fired')
        else:
            return variables



from abc import abstractmethod

from catcher.steps.step import Step
from catcher.utils.misc import fill_template_recursive


class ExternalStep(Step):
    """
    Implement this step in case you are adding external python module to
    catcher-modules project
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        method = Step.filter_predefined_keys(kwargs)
        self.data = {method: kwargs[method]}

    def simple_input(self, variables):
        """
        Use this method to get simple input as python object, with all
        templates filled in

        :param variables:
        :return: python object

        """
        return fill_template_recursive(self.data, variables)

    @abstractmethod
    def action(self, includes: dict, variables: dict) -> dict or tuple:
        """
        Perform an action.

        :param includes: Script includes.
        :param variables: Script variables.
        :return: variables and step's output. Output is optional.

        For code above
        ::

            postgres:
                request:
                    conf:
                        dbname: test
                        user: test
                        password: test
                        host: localhost
                        port: 5433
                    query: 'select count(*) from test'
            register: {documents: '{{ OUTPUT }}'}

        in_data will be
        ::

            {'request' : {'conf': {'dbname': 'test', 'user': 'test', 'password': 'test', 'host': 'localhost', 'port': 5433},
                          'query': 'select count(*) from test'}
            }

        """
        pass



import json
import os
import re
from typing import Union, Optional

import requests

from catcher.steps.step import Step, update_variables
from catcher.utils.logger import debug, warning
from catcher.utils.misc import fill_template, fill_template_str
from catcher.utils import file_utils


class Http(Step):
    """
    Perform an http request: from just getting the information from the server to pushing a file to it.

    :Input:

    :<method>: http method. Most frequent are get/post/put/delete.
     See `docs <https://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html>`_ for details

    - headers: Dictionary with custom headers. *Optional*
    - url: url to call
    - response_code: Code to await. Use 'x' for a wildcard or '-' to set a range between 2 codes.
                     *Optional* default is 200.
    - body: body to send. *Optional*.
    - body_from_file: File can be used as data source. *Optional*.
    - files: send file from resources (only for methods which support it). *Optional*
    - verify: Verify SSL Certificate in case of https. *Optional*. Default is true.
    - should_fail: true, if this request should fail, f.e. to test connection refused. Will fail the test if no errors.
    - session: http session name. Cookies are saved between sessions. *Optional*. Default session is 'default'.
               If set to null - there would be no session.
    - fix_cookies: if true will make cookies secure if you use https and not secure if you don't. *Optional*.
                   Default is true. Is useful when you don't have tls for your test env, but can't change infra.
    - timeout: number of seconds to wait for response. *Optional*. Default is no timeout (wait forever)

    :files: is a single file or list of files, where <file_param> is a name of request param.
            If you don't specify headers 'multipart/form-data' will be set automatically.

    - <file_param>: path to the file
    - type: file mime type

    :cookies: All requests are run in the session, sharing cookies got from previous requests. If you wish to
              start new empty session use `session`. If you don't want a session to be saved use `session: null`

    :Examples:

    Put data to server and await 200-299 code
    ::

        http:
          put:
            url: 'http://test.com?user_id={{ user_id }}'
            body: {'foo': bar}
            response_code: 2XX

    Put data to server and await 201-3XX code
    ::

        http:
          put:
            url: 'http://test.com?user_id={{ user_id }}'
            body: {'foo': bar}
            response_code: 201-3xx

    Post data to server with custom header
    ::

        http:
          post:
            headers: {Content-Type: 'application/json', Authorization: '{{ token }}'}
            url: 'http://test.com?user_id={{ user_id }}'
            body: {'foo': bar}

    Post file to remote server
    ::

        http:
          post:
            url: 'http://test.com'
            body_from_file: "data/answers.json"

    SSL without verification
    ::

        http:
          post:
            url: 'https://my_server.de'
            body: {'user':'test'}
            verify: false

    Manual set of json body. tojson will convert 'var' to json string
    ::

        http:
          post:
            url: 'http://test.com?user_id={{ user_id }}'
            body: '{{ var |tojson }}'

    Set json by providing json headers and passing python object to body
    ::

        http:
          post:
            url: 'http://test.com?user_id={{ user_id }}'
            headers: {Content-Type: 'application/json'}
            body: '{{ var  }}'

    Send file with a post request
    ::

        http:
          post:
            url: 'http://example.com/upload'
            files:
                file: 'subdir/my_file_in_resources.csv'
                type: 'text/csv'

    Send multiple files with a single post request
    ::

        http:
          post:
            url: 'http://example.com/upload'
            files:
                - my_csv_file: 'one.csv'
                  type: 'text/csv'
                - my_json_file: 'two.json'
                  type: 'application/json'

    Test disconnected service:
    ::

        steps:
        - docker:
            disconnect:
                hash: '{{ my_container }}'
        - http:
            get:
                url: '{{ my_container_url }}'
                should_fail: true

    Test correct and incorrect login (clear cookies):
    ::

        steps:
            - http:
                post:
                    url: 'http://test.com/login.php?user_id={{ user_id }}'
                    body: {'pwd': secret}
                    response_code: 2XX
                    session: 'user1'
                name: "Do a login"
            - http:
                get:
                    url: 'http://test.com/protected_path'
                    response_code: 2XX
                    session: 'user1'
                name: "Logged-in user can access protected_path"
            - http:
                get:
                    url: 'http://test.com/protected_path'
                    response_code: 401
                    session: 'user2'
                name: "protected_path can't be accessed without login"
    """
    sessions = {}

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        method = Step.filter_predefined_keys(kwargs)  # get/post/put...
        self.method = method.lower()
        conf = kwargs[method]
        self.url = conf['url']
        self.headers = conf.get('headers', {})
        self.verify = conf.get('verify', True)
        self._should_fail = conf.get('should_fail', False)
        if not self.verify:
            import urllib3
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        self.code = conf.get('response_code', 200)
        self.body = conf.get('body')
        self.file = conf.get('body_from_file')
        self.files = conf.get('files')
        self.session = conf.get('session', 'default')
        self.fix_cookies = conf.get('fix_cookies', True)
        self.timeout = conf.get('timeout')

    @update_variables
    def action(self, includes: dict, variables: dict) -> Union[tuple, dict]:
        url = fill_template(self.url, variables)
        session = Http.sessions.get(self.session, requests.Session())
        r = None
        try:
            r = session.request(self.method, url, **self._form_request(url, variables))
            if self._should_fail:  # fail expected
                raise RuntimeError('Request expected to fail, but it doesn\'t')
        except requests.exceptions.ConnectionError as e:
            debug(str(e))
            if self._should_fail:  # fail expected
                return variables
        self.__fix_cookies(url, session)
        if self.session is not None:  # save session if name is specified
            Http.sessions[self.session] = session
        if r is None:
            raise Exception('No response received')
        debug(r.text)
        try:
            response = r.json()
        except ValueError:
            response = r.text
        if self.__check_code(r.status_code, self.code):
            raise RuntimeError('Code mismatch: ' + str(r.status_code) + ' vs ' + str(self.code))
        return variables, response

    def _form_request(self, url, variables: dict) -> dict:
        headers = dict([(fill_template_str(k, variables), fill_template_str(v, variables))
                        for k, v in self.headers.items()])
        rq = dict(verify=self.verify, headers=headers, files=self.__form_files(variables))
        isjson, body = self.__form_body(variables)
        debug('http ' + str(self.method) + ' ' + str(url) + ', ' + str(headers) + ', ' + str(body))
        content_type = self.__get_content_type(headers)
        if isinstance(body, str):  # decode all strings to utf to prevents latin-1 errors
            body = body.encode('utf-8')
        if isjson or isinstance(body, dict):  # contains tojson or dict supplied
            if isinstance(body, dict) and content_type == 'application/json':
                # json body formed manually via python dict
                rq['json'] = body
            else:  # json string or form-data dict
                rq['data'] = body
        else:  # raw body (or body is None)
            rq['data'] = body
        rq['timeout'] = self.timeout
        return rq

    @staticmethod
    def __get_content_type(headers):
        content_type = headers.get('Content-Type')
        if content_type is None:
            content_type = headers.get('content-type')
        return content_type

    def __form_body(self, variables) -> tuple:
        if self.method == 'get':
            return False, None
        body = self.body
        if body is None and self.file is not None:
            resources = variables['RESOURCES_DIR']
            body = file_utils.read_file(fill_template_str(os.path.join(resources, self.file), variables))
        if isinstance(body, dict) or isinstance(body, list):  # dump body to json to be able fill templates in
            body = json.dumps(body)
        if body is None:
            return False, None
        isjson = 'tojson' in body  # converted to json manually via template filter
        return isjson, fill_template(body, variables, isjson=isjson)

    def __form_files(self, variables) -> Optional[list]:
        if self.files is not None:
            if isinstance(self.files, dict):
                return [self.__prepare_file(self.files, variables)]
            elif isinstance(self.files, list):
                return [self.__prepare_file(f, variables) for f in self.files]
            else:
                warning('Don\'t know how to prepare ' + type(self.files))
        return None

    def __fix_cookies(self, url: str, session):
        """
        If url was https and cookies received are not secure - make them secure.
        If url was http and cookies received are secure - make them not secure
        """
        if self.fix_cookies:
            secure = url.startswith('https')
            for site, cookies in session.cookies._cookies.items():
                for path, cookie_list in cookies.items():
                    for name, cookie in cookie_list.items():
                        cookie.secure = secure

    @staticmethod
    def __check_code(got: int, expected):
        expected_str = str(expected).lower()
        if '-' in str(expected_str):  # range
            [e_from, e_to] = expected_str.split('-')
            return not (int(e_from.replace('x', '0')) <= got <= int(e_to.replace('x', '9')))
        if 'x' in expected_str:  # regexp
            expected_str = expected_str.replace('x', '.')
        p = re.compile(expected_str)
        return p.match(str(got)) is None

    @staticmethod
    def __prepare_file(file: dict, variables: dict):
        resources = variables['RESOURCES_DIR']
        [file_key] = [k for k in file.keys() if k != 'type']
        filepath = file[file_key]
        file_type = file.get('type', 'text/plain')
        filename = file_utils.get_filename(filepath)
        file = fill_template_str(file_utils.read_file(os.path.join(resources, filepath)), variables)
        return file_key, (filename, file, file_type)



import importlib
import importlib.util
import inspect
import os
import pkgutil
import sys
from contextlib import contextmanager
from inspect import getmembers, isfunction
from os.path import join
from pydoc import locate
from types import ModuleType
from typing import Union

from catcher.utils.logger import warning, error, debug


def prepare_modules(module_paths: list, available: dict) -> dict:
    """
    Scan all paths for external modules and form key-value dict.
    :param module_paths: list of external modules (either python packages or third-party scripts)
    :param available: dict of all registered python modules (can contain python modules from module_paths)
    :return: dict of external modules, where keys are filenames (same as stepnames) and values are the paths
    """
    indexed = {}
    for path in module_paths:
        if not os.path.exists(path) and path not in available:
            err = 'No such path: ' + path
            error(err)
        else:
            for f in os.listdir(path):
                mod_path = join(path, f)
                if f in indexed:
                    warning('Override ' + indexed[f] + ' with ' + mod_path)
                indexed[f] = mod_path
    return indexed


def get_submodules_of(package: str):
    """
    Get all submodules and their importers for the package. It is not recursive.
    For recursive see __load_python_package_installed
    Package should be installed in the system.
    """
    modules = locate(package)
    return [(modname, importer) for importer, modname, ispkg
            in pkgutil.iter_modules(path=modules.__path__, prefix=modules.__name__ + '.')]


def load_external_actions(package: str):
    """
    Load all classes from a package
    """
    if package.endswith('.py'):
        return __load_python_package_by_path(package)
    else:
        return __load_python_package_installed(package)


def get_all_subclasses_of(clazz) -> list:
    return clazz.__subclasses__() + [g for s in clazz.__subclasses__()
                                     for g in get_all_subclasses_of(s)]


def is_package_installed(package: str) -> bool:
    try:
        importlib.import_module(package)
        return True
    except ImportError:
        return False


def add_package_to_globals(package: str, glob=None, warn_missing_package=True) -> dict:
    if glob is None:
        glob = globals()
    try:
        mod = importlib.import_module(package)
        glob[package] = mod
    except ImportError as e:
        if warn_missing_package:
            warning(str(e))
        else:
            debug(str(e))
    return glob


def get_all_functions(module: str) -> dict:
    mod = load_external_actions(module)
    if isinstance(mod, list):
        res = {}
        for m in mod:
            res = {**res, **dict([o for o in getmembers(m) if isfunction(o[1])])}
        return res
    else:
        return dict([o for o in getmembers(mod) if isfunction(o[1])])


def get_all_classes(module: Union[str, ModuleType]) -> dict:
    if isinstance(module, str):
        module = sys.modules[module]
    return dict(inspect.getmembers(module, inspect.isclass))


@contextmanager
def add_to_path(p):
    import sys
    old_path = sys.path
    sys.path = sys.path[:]
    sys.path.insert(0, p)
    try:
        yield
    finally:
        sys.path = old_path


def __load_python_package_by_path(path: str):
    with add_to_path(os.path.dirname(path)):
        spec = importlib.util.spec_from_file_location(path, path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module


def __load_python_package_installed(package: str):
    modules = locate(package)
    if modules is None:
        return  # package not installed
    if not hasattr(modules, '__path__'):
        return modules
    return [importlib.import_module(modname)
            for importer, modname, ispkg in pkgutil.walk_packages(path=modules.__path__,
                                                                  prefix=modules.__name__ + '.',
                                                                  onerror=lambda x: None)]



import os
import subprocess
from os.path import join

from test import TEST_DIR
from test.abs_test_class import TestClass
from catcher.utils import file_utils
from catcher import __main__


class E2ETest(TestClass):
    def __init__(self, method_name):
        super().__init__('e2e_test', method_name)

    def setUp(self):
        super().setUp()
        file_utils.ensure_empty(join(os.getcwd(), TEST_DIR, 'steps'))

    def tearDown(self):
        super().tearDown()
        file_utils.remove_dir(join(os.getcwd(), TEST_DIR, 'steps'))
        file_utils.remove_dir(join(os.getcwd(), TEST_DIR, 'reports'))

    def test_can_run(self):
        self.populate_file('main.yaml', '''---
                steps:
                    - check: 
                        equals: {the: 'life', is: 'life'}  # na-na, na-na-na
                ''')
        self._run_test(self.test_dir)

    def test_check_output(self):
        self.populate_file('one.yaml', '''---
                        steps:
                            - echo: {from: '{{ "test" | hash }}'}
                            - check: 
                                equals: {the: 'life', is: 'life'}
                        ''')
        self.populate_file('two.yaml', '''---
                        steps:
                            - echo: {from: '{{ "test" | hash }}'}
                            - check: 
                                equals: {the: 'life', is_not: 'life'}
                            - echo: 'hello world'
                        ''')
        self.populate_file('three.yaml', '''---
                        steps:
                            - echo: {from: '{{ "test" | hash }}'}
                            - check: 
                                equals: {the: 'life', is: 'life'}
                        ''')
        output = self._run_test(self.test_dir + ' --no-color', expected_code=1)
        lines = output.strip().split('\n')
        to_compare = sorted(lines[-3:])  # need to sort it, as output order is not guaranteed in CI
        self.assertEqual('INFO:catcher:Test run 3. Success: 2, Fail: 1. Total: 67%', lines[-4])
        self.assertEqual('Test one: pass', to_compare[-3])
        self.assertEqual('Test three: pass', to_compare[-2])
        self.assertEqual('Test two: fail, on step 2', to_compare[-1])

    def test_check_output_skipp_test(self):
        self.populate_file('one.yaml', '''---
                                ignore: true
                                steps:
                                    - echo: {from: '{{ "test" | hash }}'}
                                    - check: 
                                        equals: {the: 'life', is: 'life'}
                                ''')
        self.populate_file('two.yaml', '''---
                                steps:
                                    - echo: {from: '{{ "test" | hash }}'}
                                    - check: 
                                        equals: {the: 'life', is: 'life'}
                                    - echo: 'hello world'
                                ''')
        output = self._run_test(self.test_dir + ' --no-color')
        lines = output.strip().split('\n')
        to_compare = sorted(lines[-3:])  # need to sort it, as output order is not guaranteed in CI
        self.assertEqual('INFO:catcher:Test run 2. Success: 1, Fail: 0, Skipped: 1. Total: 100%', lines[-3])
        self.assertEqual('Test one: skipped', to_compare[-2])
        self.assertEqual('Test two: pass', to_compare[-1])

    def test_check_output_run_on_include(self):
        self.populate_step('steps/include.yaml', '''---
                                steps:
                                    - echo: {from: 'I am include'}
                                    - check: 
                                        equals: {the: 'life', is: 'life'}
                                ''')
        self.populate_file('two.yaml', '''---
                                include: steps/include.yaml
                                steps:
                                    - echo: {from: '{{ "test" | hash }}'}
                                    - check: 
                                        equals: {the: 'life', is_not: 'life'}
                                    - echo: 'hello world'
                                ''')
        self.populate_file('three.yaml', '''---
                                steps:
                                    - echo: {from: '{{ "test" | hash }}'}
                                    - check: 
                                        equals: {the: 'life', is: 'life'}
                                ''')
        output = self._run_test(self.test_dir + ' --no-color', expected_code=1)
        lines = output.strip().split('\n')
        to_compare = sorted(lines[-3:])  # need to sort it, as output order is not guaranteed in CI
        self.assertEqual('INFO:catcher:Test run 2. Success: 1, Fail: 1. Total: 50%', to_compare[-3])
        self.assertEqual('Test three: pass', to_compare[-2])
        self.assertEqual('Test two: fail, on step 2', to_compare[-1])

    def test_check_output_ignored_include(self):
        self.populate_step('steps/include.yaml', '''---
                                        ignore: true
                                        steps:
                                            - echo: {from: 'I am include'}
                                            - check: 
                                                equals: {the: 'life', is: 'life'}
                                        ''')
        self.populate_file('two.yaml', '''---
                                        include: steps/include.yaml
                                        steps:
                                            - echo: {from: '{{ "test" | hash }}'}
                                            - check: 
                                                equals: {the: 'life', is_not: 'life'}
                                            - echo: 'hello world'
                                        ''')
        self.populate_file('three.yaml', '''---
                                        steps:
                                            - echo: {from: '{{ "test" | hash }}'}
                                            - check: 
                                                equals: {the: 'life', is: 'life'}
                                        ''')
        output = self._run_test(self.test_dir + ' --no-color', expected_code=1)
        lines = output.strip().split('\n')
        to_compare = sorted(lines[-3:])  # need to sort it, as output order is not guaranteed in CI
        self.assertEqual('INFO:catcher:Test run 2. Success: 1, Fail: 1. Total: 50%', to_compare[-3])
        self.assertEqual('Test three: pass', to_compare[-2])
        self.assertEqual('Test two: fail, on step 2', to_compare[-1])

    def test_failed_include_ignore_test(self):
        self.populate_file('two.yaml', '''---
                                        include: non_existent.yaml
                                        ignore: true
                                        steps:
                                            - echo: 'hello world'
                                        ''')
        output = self._run_test(self.test_dir + ' --no-color')
        print(output)
        lines = output.strip().split('\n')
        to_compare = sorted(lines[-3:])  # need to sort it, as output order is not guaranteed in CI
        self.assertEqual('INFO:catcher:Test run 1. Success: 0, Fail: 0, Skipped: 1. Total: 100%', to_compare[0])
        self.assertEqual('INFO:catcher:Test two skipped.', to_compare[1])

    def test_check_output_run_on_action(self):
        self.populate_step('steps/include.yaml', '''---
                                        steps:
                                            - echo: {from: 'I am include'}
                                            - check: 
                                                equals: {the: 'life', is: 'life'}
                                        ''')
        self.populate_file('two.yaml', '''---
                                        include: 
                                            - file: steps/include.yaml
                                              as: run_me
                                        steps:
                                            - run:
                                                include: run_me
                                            - check: 
                                                equals: {the: 'life', is_not: 'life'}
                                            - echo: 'hello world'
                                        ''')
        self.populate_file('three.yaml', '''---
                                        steps:
                                            - echo: {from: '{{ "test" | hash }}'}
                                            - check: 
                                                equals: {the: 'life', is: 'life'}
                                        ''')
        output = self._run_test(self.test_dir + ' --no-color', expected_code=1)
        lines = output.strip().split('\n')
        to_compare = sorted(lines[-3:])  # need to sort it, as output order is not guaranteed in CI
        self.assertEqual('INFO:catcher:Test run 2. Success: 1, Fail: 1. Total: 50%', to_compare[-3])
        self.assertEqual('Test three: pass', to_compare[-2])
        self.assertEqual('Test two: fail, on step 2', to_compare[-1])

    def test_check_output_ignored_run_on_action(self):
        self.populate_step('steps/include.yaml', '''---
                                        ignore: true
                                        steps:
                                            - echo: {from: 'I am include'}
                                            - check: 
                                                equals: {the: 'life', is: 'life'}
                                        ''')
        self.populate_file('two.yaml', '''---
                                        include: 
                                            - file: steps/include.yaml
                                              as: run_me
                                        steps:
                                            - run:
                                                include: run_me
                                            - check: 
                                                equals: {the: 'life', is_not: 'life'}
                                            - echo: 'hello world'
                                        ''')
        self.populate_file('three.yaml', '''---
                                        steps:
                                            - echo: {from: '{{ "test" | hash }}'}
                                            - check: 
                                                equals: {the: 'life', is: 'life'}
                                        ''')
        output = self._run_test(self.test_dir + ' --no-color', expected_code=1)
        lines = output.strip().split('\n')
        to_compare = sorted(lines[-3:])  # need to sort it, as output order is not guaranteed in CI
        self.assertEqual('INFO:catcher:Test run 2. Success: 1, Fail: 1. Total: 50%',
                         to_compare[-3])
        self.assertEqual('Test three: pass', to_compare[-2])
        self.assertEqual('Test two: fail, on step 2', to_compare[-1])

    def test_run_summary_with_output(self):
        self.populate_file('main.yaml', '''---
                        steps:
                            - check: 
                                equals: {the: 'life', is: 'life'}  # na-na, na-na-na
                        ''')
        output = self._run_test(self.test_dir + ' --no-color -p json')
        lines = output.strip().split('\n')
        to_compare = sorted(lines[-2:])  # need to sort it, as output order is not guaranteed in CI
        self.assertEqual('INFO:catcher:Test run 1. Success: 1, Fail: 0. Total: 100%', to_compare[-2])
        self.assertEqual('Test main: pass', to_compare[-1])

    def test_run_output_include_only(self):
        self.populate_file('main.yaml', '''---
                include: 
                    - steps/simple_file.yaml
                ''')
        self.populate_step('steps/simple_file.yaml', '''---
                include: 
                    - steps/other_simple_file.yaml
                variables:
                    foo: bar
                steps:
                    - echo: {from: '{{ foo }}'}
                ''')
        self.populate_step('steps/other_simple_file.yaml', '''---
                include: 
                    - steps/simple_file.yaml
                variables:
                    foo: baz
                steps:
                    - echo: {from: '{{ foo }}'}
                ''')
        output = self._run_test(self.test_dir + ' --no-color', expected_code=1)
        lines = output.strip().split('\n')
        self.assertTrue('Circular dependencies' in lines[-3])
        self.assertEqual('INFO:catcher:Test run 1. Success: 0, Fail: 1. Total: 0%', lines[-2])
        self.assertEqual('Test main: fail, on step 0', lines[-1])

    def test_run_output_limited(self):
        pass

    def test_run_output_final(self):
        pass

    def _run_test(self, args: str, expected_code=0):
        process = subprocess.Popen('python {} {}'.format(__main__.__file__, args).split(' '),
                                   cwd=join(os.getcwd(), TEST_DIR),
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE,
                                   universal_newlines=True)
        stdout, stderr = process.communicate()
        if expected_code != process.returncode:
            print(stderr)
            self.fail('Got return code {}, expected {}'.format(process.returncode, expected_code))
        return stderr

    @staticmethod
    def populate_step(file: str, content: str):
        with open(join(os.getcwd(), TEST_DIR, file), 'w') as f:
            f.write(content)



import os
import traceback
from os.path import join

from catcher.core.var_holder import VariablesHolder
from catcher.core.parser import Parser
from catcher.core.step_factory import StepFactory
from catcher.core.test import Test
from catcher.core.filters_factory import FiltersFactory
from catcher.modules.log_storage import LogStorage
from catcher.steps.step import SkipException
from catcher.utils import logger
from catcher.utils.file_utils import cut_path
from catcher.utils.logger import warning, info, debug, OptionalOutput
from catcher.core.mod_factory import ModulesFactory


class Runner:
    def __init__(self,
                 path: str,
                 tests_path: str,
                 inventory: str or None,
                 modules=None,
                 cmd_env=None,
                 system_environment=None,
                 resources=None,
                 output_format=None,
                 filter_list=None) -> None:
        # singletons init should be done before services (like vars holder), as singletons maybe used there
        # modules should be done before filters, as requirements module installs dependencies for custom bifs
        ModulesFactory(resources_dir=resources or os.path.join(path, 'resources'))
        FiltersFactory(custom_modules=filter_list)
        StepFactory(modules)

        self.tests_path = tests_path
        self.path = path
        self.parser = Parser(path, inventory)
        self.var_holder = VariablesHolder(path,
                                          system_environment=system_environment,
                                          inventory_vars=self.parser.read_inventory(),
                                          cmd_env=cmd_env,
                                          resources=resources)
        if output_format:
            logger.log_storage = LogStorage(output_format)

    def run_tests(self, output: str = 'full') -> bool:
        """
        Run the testcase
        :param output: 'full' - all output possible. 'limited' - only test end report & summary. 'final' - only summary.
        """
        try:
            [mod.before() for mod in ModulesFactory().modules.values()]
            results = []
            for parse_result in self.parser.read_tests(self.tests_path):
                if parse_result.should_run:  # parse successful
                    variables = self.var_holder.variables  # each test has it's own copy of global variables
                    variables['TEST_NAME'] = parse_result.test.file  # variables are shared between test and includes
                    with OptionalOutput(output == 'final'):
                        for include in parse_result.run_on_include:  # run all includes before the main test.
                            self._run_test(include, variables, output=output, test_type='include')
                        result = self._run_test(parse_result.test, variables, output=output)
                        results.append(result)
                        self._run_finally(parse_result.test, result)
                else:  # parse failed (dependency/parsing problem)
                    warning('Test ' + cut_path(self.tests_path, parse_result.test) +
                            logger.red(' failed: ') + str(parse_result.parse_error))
                    logger.log_storage.test_parse_fail(parse_result.test, parse_result.parse_error)
                    results.append(False)
            return all(results)
        finally:
            logger.log_storage.write_report(join(self.path, 'reports'))
            logger.log_storage.print_summary(self.tests_path)
            [mod.after() for mod in ModulesFactory().modules.values()]

    def _run_test(self, test: Test, global_variables: dict, output: str = 'full', test_type='test') -> bool:
        try:
            self.var_holder.prepare_variables(test, global_variables)
            logger.log_storage.test_start(test.file, test_type=test_type)
            test.check_ignored()
            with OptionalOutput(output == 'limited'):
                test.run()
            info(test_type.capitalize() + ' ' + cut_path(self.tests_path, test.file) + logger.green(' passed.'))
            logger.log_storage.test_end(test.file, True, test_type=test_type)
            return True
        except SkipException:
            info(test_type.capitalize() + ' ' + cut_path(self.tests_path, test.file) + logger.yellow(' skipped.'))
            logger.log_storage.test_end(test.file, True, end_comment='Skipped', test_type=test_type)
            return True
        except Exception as e:
            warning(test_type.capitalize() + ' ' + cut_path(self.tests_path, test.file) +
                    logger.red(' failed: ') + str(e))
            debug(traceback.format_exc())
            logger.log_storage.test_end(test.file, False, str(e), test_type=test_type)
            return False

    def _run_finally(self, test, result: bool):
        if test and test.final:
            logger.log_storage.test_start(test.file, test_type='{}_cleanup'.format(test.file))
            try:
                test.run_finally(result)
                info('Test ' + cut_path(self.tests_path, test.file) + ' [cleanup] ' + logger.green(' passed.'))
                logger.log_storage.test_end(test.file, True, test_type='{} [cleanup]'.format(test.file))
            except Exception as e:
                warning('Test ' + cut_path(self.tests_path, test.file) + ' [cleanup] ' +
                        logger.red(' failed: ') + str(e))
                debug(traceback.format_exc())
                logger.log_storage.test_end(test.file, False, test_type='{} [cleanup]'.format(test.file))



import os

import networkx as nx

from typing import Tuple, Optional, Union, List

from catcher.utils.logger import debug


class Include:
    """
    Include another testcase in include section:

    - file: test file to include
    - variables: variables to override
    - alias: unique name to run this test later
    - run_on_include: should run as soon as included (before test)
    - ignore_errors: should continue running other tests if this fails

    :Examples:

    Simple include - will run before test
    ::

        include: register_user.yaml

    Will run before test with some variables overridden
    ::

        include:
            file: register_user.yaml
            variables:
                user_email: 'override@email.org'

    Multiple includes. Will run one by one. Variables from the first one will be passed to the next
    ::

        include:
            - simple_form.yaml
            - file: long_form.yaml
              variables: {user_email: 'override@email.org'}

    Include with alias and run later
    ::

        include:
            file: register_user.yaml
            as: sign_up
        steps:
            # .... some steps
            - run:
                include: sign_up
            # .... some steps

    See `includes <https://catcher-test-tool.readthedocs.io/en/latest/source/includes.html>`_ for more info.
    """

    def __init__(self, ignore_errors=False, file=None, **kwargs) -> None:
        if 'file' is None:
            raise RuntimeError('Can\'t include unknown file.')
        self.file = file
        self.variables = kwargs.get('variables', {})
        self.alias = kwargs.get('as', None)
        self.run_on_include = kwargs.get('run_on_include', self.alias is None)
        self.ignore_errors = ignore_errors

    @staticmethod
    def check_circular(parent: str,
                       all_includes: nx.DiGraph,
                       current_include: dict) -> 'Include':
        path = current_include['file']
        all_includes.add_edge(parent, path)
        if list(nx.simple_cycles(all_includes)):
            debug(str(all_includes.edges))
            raise Exception('Circular dependencies for ' + path)
        return Include(**current_include)

    @staticmethod
    def read_includes(path: str,
                      parent: str,
                      includes: Union[dict, list, str],
                      all_includes: Optional[nx.DiGraph]) -> Tuple[nx.DiGraph, List['Include']]:
        if all_includes is None:
            all_includes = nx.DiGraph()
        if isinstance(includes, str) or isinstance(includes, dict):  # single include
            raw = [Include.read_include(path, parent, includes, all_includes)]
        elif isinstance(includes, list):  # an array of includes
            raw = [Include.read_include(path, parent, include, all_includes) for include in includes]
        else:
            raw = []
        return all_includes, raw

    @staticmethod
    def read_include(path=None, parent=None, include_file=None, all_includes=None):
        include_file = Include.path_from_root(include_file, path)
        return Include.check_circular(parent, all_includes, include_file)

    @staticmethod
    def path_from_root(include_file: str or dict, path: str) -> dict:
        if isinstance(include_file, str):
            return {'file': os.path.join(path, include_file)}
        else:
            include_file['file'] = os.path.join(path, include_file['file'])
            return include_file



import base64
import hashlib
import random
import sys
import datetime
import time

from faker import Faker
from catcher.utils import module_utils, misc

random.seed()


def filter_b64encode(param):
    """
    Encode string as base64
    F.e. ::

        - http:
              post:
                url: '{{ auth_service }}/oauth/token'
                headers: {Authorization: '{{ admin_token | b64encode }}'}

    :param param: data to convert
    """
    return base64.b64encode(param.encode('utf-8')).decode('utf-8')


def filter_b64decode(param):
    """
        Decode base64 encoded string
        F.e. ::

            - echo: '{{ token |b64decode }}'

        :param param: data to convert
        """
    return base64.b64decode(param.encode('utf-8')).decode('utf-8')


def filter_astuple(param):
    """
    Convert data to tuple
    F.e. ::

        - postgres:
            request:
              conf: '{{ postgres_conf }}'
              query: "select status from my_table
                     where id in {{ [1, 2, 3] |astuple }}"

    :param param: data to convert
    """
    return tuple(misc.try_get_objects(param))


def filter_asint(param):
    """
    Convert data to int
    F.e. ::

        - postgres:
            request:
              conf: '{{ postgres_conf }}'
              query: "select status from my_table
                     where id == {{ my_str_var |asint }}"

    :param param: data to convert
    """
    return int(misc.try_get_objects(param))


def filter_asfloat(param):
    """
    Convert data to float
    F.e. ::

        - check: {equals: {the: 36.6, is: '{{ "36.6" | asfloat }}'}}

    :param param: data to convert
    """
    return float(misc.try_get_objects(param))


def filter_aslist(param):
    """
    Convert data to list
    F.e. ::

        - loop:
            foreach:
                in: '{{ my_dictionary |aslist }}'
                do:
                    echo: {from: '{{ ITEM.value }}', to: '{{ ITEM.key }}.output'}

    :param param: data to convert
    """
    return list(misc.try_get_objects(param))


def filter_asdict(param):
    """
    Convert data to dict
    F.e. ::

        - check: {equals: {the: [1, 2],
                           is: '{{ ([("one", 1), ("two", 2)] | asdict).values() |aslist }}'}}

    :param param: data to convert
    """
    return dict(misc.try_get_objects(param))


def filter_asstr(param):
    """
    Convert data to string
    F.e. ::

        - check: {equals: {the: '17', is: '{{ my_int | asstr }}'}}

    :param param: data to convert
    """
    return str(misc.try_get_objects(param))


def function_random(param, locale=None):
    """
    Call `Faker <https://github.com/joke2k/faker>`_ and return it's result. Is used to generate random data.
    F.e. ::

        - echo: {from: '{{ random("email") }}', to: one.output}

    :param param: Faker's provider name.
    :param locale: Faker's locale param
    """
    fake = Faker(locale=locale)
    for modname, importer in module_utils.get_submodules_of('faker.providers'):  # add all known providers
        fake.add_provider(importer.find_module(modname).load_module(modname))
    if hasattr(fake, param):
        return getattr(fake, param)()
    else:
        raise ValueError('Unknown param to randomize: ' + param)


def filter_hash(data, alg='md5'):
    """
    Filter for hashing data.
    F.e. ::

        - echo: {from: '{{ my_var | hash("sha1") }}', to: two.output}

    :param data: data to hash
    :param alg: algorithm to use
    """
    if hasattr(hashlib, alg):
        m = getattr(hashlib, alg)()
        m.update(data.encode())
        return m.hexdigest()
    else:
        raise ValueError('Unknown algorithm: ' + data)


def function_now(date_format='%Y-%m-%d %H:%M:%S.%f'):
    """
    Get current date in a specified format.
    F.e. ::

        - echo: {from: '{{ now("%Y-%m-%d") }}', to: year.output}
    :param date_format: date format
    """
    return datetime.datetime.now().strftime(date_format)


def function_now_ts():
    """
    Get current date time in as a timestamp.
    F.e. ::

        - echo: {from: '{{ now_ts() }}', to: timestamp.output}
    """
    return round(time.time(), 6)  # from timestamp uses rounding, so we should also use it here, to make them compatible


def filter_astimestamp(data, date_format='%Y-%m-%d %H:%M:%S.%f'):
    """
    Convert date to timestamp. Date can be either python date object or date string
    F.e. ::

        - echo: {from: '{{ date_time_var | astimestamp }}', to: two.output}

    :param data: date time object (or string representation) to be converted to a timestamp.
    :param date_format: date format (in case it is a string)
    """
    if isinstance(data, str):
        data = datetime.datetime.strptime(data, date_format)
    return datetime.datetime.timestamp(data)


def filter_asdate(data, date_format='%Y-%m-%d %H:%M:%S.%f'):
    """
    Convert timestamp to date
    F.e. ::

        - echo: {from: '{{ timestamp_var | asdate(date_format="%Y-%m-%d") }}', to: two.output}

    :param data: timestamp to be converted to a date
    :param date_format: expected data format.
    """
    if isinstance(data, str):
        if '.' in data:
            data = float(data)
        else:
            data = int(data)
    return datetime.datetime.fromtimestamp(data).strftime(date_format)


def function_random_int(range_from=-sys.maxsize - 1, range_to=sys.maxsize):
    """
    Function for random number return. Output can be controlled by `range_from` and `range_to` attributes.
    F.e. ::

        - echo: {from: '{{ random_int(range_from=1) }}', to: one.output}

    """
    return random.randint(range_from, range_to)


def function_random_choice(sequence):
    """
    Function to make a random choice in a collection.
    F.e. ::

        - echo: {from: '{{ random_choice([1, 2, 3]) }}', to: one.output}

    :param sequence: collection of elements to choose from.
    """
    return random.choice(sequence)
